/*
Module : cpdh.h
Purpose: Defines the interface for a C++ wrapper class for the PDH API

Copyright (c) 1999 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


///////////////////////////////// Macros / Defines ////////////////////////////

#pragma once

#ifndef __CPDH_H__
#define __CPDH_H__

#ifndef _PDH_H_
#pragma message("To avoid this message, please put pdh.h in your pre compiled header (normally stdafx.h)")
#include <pdh.h>
#endif //#ifndef _PDH_H_

#ifndef _EXCEPTION_
#pragma message("To avoid this message, please put exception in your pre compiled header (normally stdafx.h)")
#include <exception>
#endif //#ifndef _EXCEPTION_

#ifndef _STRING_
#pragma message("To avoid this message, please put string in your pre compiled header (normally stdafx.h)")
#include <string>
#endif //#ifndef _STRING_

#ifndef _VECTOR_
#pragma message("To avoid this message, please put vector in your pre compiled header (normally stdafx.h)")
#include <vector>
#endif //#ifndef _VECTOR_

#ifdef _UNICODE
  using CPDHString = std::wstring;
#else
  using CPDHString = std::string;
#endif //#ifdef _UNICODE
using CPDHStringArray = std::vector<CPDHString>;


///////////////////////////////// Classes /////////////////////////////////////

//Represents an exception condition relating to a PDH classes as below
class CPdhException : public std::exception
{
public:
//Constructor / Destructors
  CPdhException() = delete;
  CPdhException(_In_ PDH_STATUS Status) noexcept;
  CPdhException(const CPdhException&) = default;
  CPdhException(CPdhException&&) = default;
  ~CPdhException() = default;

//Member variables
  PDH_STATUS m_Status;

//Methods
  CPdhException& operator=(const CPdhException&) = default;
  CPdhException& operator=(CPdhException&&) = default;
  BOOL GetErrorMessage(_Out_writes_z_(nMaxError) LPTSTR lpszError, _In_ UINT nMaxError, _Out_opt_ PUINT pnHelpContext = nullptr);
#pragma warning(suppress: 26495)
};


//Info relating to a PDH counter as returned in CPdhCounter::GetInfo
class CPdhCounterInfo
{
public:
//Constructors / Destructors
  CPdhCounterInfo() noexcept : m_dwType(0),
                               m_CVersion(0),
                               m_CStatus(0),
                               m_lScale(0),
                               m_lDefaultScale(0),
                               m_dwUserData(0),
                               m_dwQueryUserData(0),
                               m_dwInstanceIndex(0)
  {
  }
  CPdhCounterInfo(const CPdhCounterInfo&) = delete;
  CPdhCounterInfo(CPdhCounterInfo&&) = delete;
  ~CPdhCounterInfo() = default;

//Methods
  CPdhCounterInfo& operator=(const CPdhCounterInfo&) = delete;
  CPdhCounterInfo& operator=(CPdhCounterInfo&&) = delete;

//Member variables
  DWORD m_dwType;
  DWORD m_CVersion;
  DWORD m_CStatus;
  LONG m_lScale;
  LONG m_lDefaultScale;
  DWORD_PTR m_dwUserData;
  DWORD_PTR m_dwQueryUserData;
  CPDHString m_sFullPath;
  CPDHString m_sMachineName;
  CPDHString m_sInstanceName;
  CPDHString m_sParentInstance;
  DWORD m_dwInstanceIndex;
  CPDHString m_sCounterName;
  CPDHString m_sExplainText;
};


//Class which encapsulates a PDH counter
class CPdhCounter
{
public:
//Constructors / Destructors
  CPdhCounter(_In_ LPCTSTR pszFullCounterPath, _In_ DWORD_PTR dwUserData = 0);
  CPdhCounter(_In_ PDH_COUNTER_PATH_ELEMENTS& CounterPathElements, _In_ DWORD_PTR dwUserData = 0);
  CPdhCounter() = delete;
  CPdhCounter(const CPdhCounter&) = delete;
  CPdhCounter(CPdhCounter&&) = delete;
  virtual ~CPdhCounter() {}; //NOLINT(modernize-use-equals-default)

//Methods
  CPdhCounter& operator=(const CPdhCounter&) = delete;
  CPdhCounter& operator=(CPdhCounter&&) = delete;
  DWORD GetRawValue(_Out_ PDH_RAW_COUNTER& value);
  void SplitPath(_Inout_ CPDHString& sMachineName, _Inout_ CPDHString& sObjectName, _Inout_ CPDHString& sInstanceName, _Inout_ CPDHString& sParentInstance, _Inout_ DWORD& dwInstanceIndex, _Inout_ CPDHString& sCounterName);
  void GetInfo(_Inout_ CPdhCounterInfo& info);
  void CalculateFromRawValue(_In_ DWORD dwFormat, _In_ PDH_RAW_COUNTER& rawValue1, _In_ PDH_RAW_COUNTER& rawValue2, _Out_ PDH_FMT_COUNTERVALUE& fmtValue);
  DWORD GetFormattedValue(_In_ DWORD dwFormat, _Out_ PDH_FMT_COUNTERVALUE& value);
  void ComputeStatistics(_In_ DWORD dwFormat, _In_ DWORD dwFirstEntry, _In_ DWORD dwNumEntries, _In_ PPDH_RAW_COUNTER lpRawValueArray, _Out_ PPDH_STATISTICS data);
  void SetScaleFactor(_In_ LONG lFactor);
  _NODISCARD LONGLONG GetTimeBase();
  _NODISCARD PPDH_FMT_COUNTERVALUE_ITEM GetFormattedCounterArray(_In_ DWORD dwFormat, _Out_ DWORD& dwItemsInArray);
  _NODISCARD PPDH_RAW_COUNTER_ITEM GetRawCounterArray(_Out_ DWORD& dwItemsInArray);
  operator HCOUNTER() noexcept { return m_hCounter; };
  _NODISCARD virtual LPVOID AllocateMemory(_In_ DWORD dwBytes) noexcept;
  virtual void DeallocateMemory(__drv_freesMem(Mem) _Frees_ptr_opt_ void* pData) noexcept;

protected:
//Member variables
  PDH_HCOUNTER m_hCounter;
  CPDHString m_sCounterPath;
  DWORD_PTR m_dwUserData;

  friend class CPdhQuery;
};


//Class which encapsulates a PDH query
class CPdhQuery
{
public:
//Constructors / Destructors
  CPdhQuery() noexcept;
  CPdhQuery(const CPdhQuery&) = delete;
  CPdhQuery(CPdhQuery&&) = delete;
  ~CPdhQuery();

//Methods
  CPdhQuery& operator=(const CPdhQuery&) = delete;
  CPdhQuery& operator=(CPdhQuery&&) = delete;
  _NODISCARD BOOL IsRealTime() noexcept;
  void Open(_In_opt_ LPCTSTR szDataSource, _In_ DWORD_PTR dwUserData = 0);
  void OpenH(_In_opt_ PDH_HLOG hDataSource, _In_ DWORD_PTR dwUserData = 0);
  void Close() noexcept;
  void Add(_Inout_ CPdhCounter& counter, _In_ BOOL bUseEnglishName = FALSE);
  void Remove(_Inout_ CPdhCounter& counter);
  void Collect();
  void Collect(_In_ DWORD dwIntervalTime, _In_ HANDLE hNewDataEvent);
  LONGLONG CollectWithTime();
  void SetTimeRange(_In_ PPDH_TIME_INFO pInfo);
  operator HQUERY() noexcept { return m_hQuery; };

protected:
//Member variables
  PDH_HQUERY m_hQuery;
  DWORD m_dwCountersInThisQuery;
};


//Class which encapasulates a PDH data source
class CPdhDataSource
{
public:
//Constructors / Destructors
  CPdhDataSource() noexcept;
  CPdhDataSource(const CPdhDataSource&) = delete;
  CPdhDataSource(CPdhDataSource&&) = delete;
  virtual ~CPdhDataSource();

//Methods
  CPdhDataSource&  operator=(const CPdhDataSource&) = delete;
  CPdhDataSource& operator=(CPdhDataSource&&) = delete;
  void Open(_In_ const CPDHStringArray& LogFileNames);
  void Open(_In_ LPCTSTR szLogFileName, _In_ DWORD dwAccessFlags, _Inout_ LPDWORD lpdwLogType, _In_opt_ PDH_HQUERY hQuery, _In_ DWORD dwMaxSize, _In_opt_ LPCTSTR szUserCaption);
  void Close(_In_ DWORD dwFlags = 0) noexcept;
  void Update(_In_ LPCTSTR szUserString);
  _NODISCARD LONGLONG GetSize();
  _NODISCARD PPDH_RAW_LOG_RECORD ReadRawRecord(_In_ FILETIME ftRecord);
  operator PDH_HLOG() noexcept { return m_hDataSource; };
  _NODISCARD CPDHString GetDefaultPerfCounter(_In_opt_ LPCTSTR szMachineName, _In_ LPCTSTR szObjectName);
  void UpdateLogFileCatalog();
  _NODISCARD virtual LPVOID AllocateMemory(_In_ DWORD dwBytes) noexcept;
  virtual void DeallocateMemory(__drv_freesMem(Mem) _Frees_ptr_opt_ void* pData) noexcept;

protected:
//Member variables
  PDH_HLOG m_hDataSource;
};


//Other PDH functions which don't belong anywhere else
class CPdh
{
public:
//Static Methods
  static void ValidatePath(_In_ LPCTSTR pszFullCounterPath);
  static void ValidatePath(_In_opt_ PDH_HLOG hDataSource, _In_ LPCTSTR szFullPathBuffer);
  _NODISCARD static CPDHStringArray ExpandPath(_In_ LPCTSTR pszWildCardPath);
  _NODISCARD static CPDHStringArray ExpandWildCardPath(_In_opt_ LPCTSTR szDataSource, _In_ LPCTSTR szWildCardPath, _In_ DWORD dwFlags);
  _NODISCARD static CPDHStringArray ExpandWildCardPathH(_In_opt_ PDH_HLOG hDataSource, _In_ LPCTSTR szWildCardPath, _In_ DWORD dwFlags);
  static void ConnectMachine(_In_opt_ LPCTSTR pszMachineName);
  _NODISCARD static CPDHStringArray EnumMachines(_In_opt_ LPCTSTR pszDataSource);
  _NODISCARD static CPDHStringArray EnumMachinesH(_In_opt_ PDH_HLOG hDataSource);
  _NODISCARD static CPDHStringArray EnumObjects(_In_opt_ LPCTSTR pszDataSource, _In_opt_ LPCTSTR pszMachineName, _In_ DWORD dwDetailLevel, _In_ BOOL bRefresh);
  _NODISCARD static CPDHStringArray EnumObjectsH(_In_opt_ PDH_HLOG hDataSource, _In_opt_ LPCTSTR pszMachineName, _In_ DWORD dwDetailLevel, _In_ BOOL bRefresh);
  static void EnumObjectItems(_In_opt_ LPCTSTR szDataSource, _In_opt_ LPCTSTR szMachineName, _In_ LPCTSTR szObjectName, _Out_ CPDHStringArray& counters, _Out_ CPDHStringArray& instances, _In_ DWORD dwDetailLevel, _In_ DWORD dwFlags = 0);
  static void EnumObjectItemsH(_In_opt_ PDH_HLOG hDataSource, _In_opt_ LPCTSTR szMachineName, _In_ LPCTSTR szObjectName, _Out_ CPDHStringArray& counters, _Out_ CPDHStringArray& instances, _In_ DWORD dwDetailLevel, _In_ DWORD dwFlags = 0);
  static void GetDefaultCounter(_In_opt_ LPCTSTR pszMachineName, _In_ LPCTSTR pszObjectName, _Inout_ CPDHString& sDefaultCounterName);
  static void GetDefaultObject(_In_opt_ LPCTSTR pszDataSource, _In_opt_ LPCTSTR pszMachineName, _Inout_ CPDHString& sDefaultObjectName);
  static void GetDefaultObjectH(_In_opt_ PDH_HLOG hDataSource, _In_opt_ LPCTSTR pszMachineName, _Inout_ CPDHString& sDefaultObjectName);
  static void BrowseCounters(_In_ PDH_BROWSE_DLG_CONFIG& BrowseDlgData);
  static void BrowseCounters(_In_ PDH_BROWSE_DLG_CONFIG_H& BrowseDlgData);
  _NODISCARD static CPDHString SelectDataSource(_In_ HWND hWndOwner, _In_ DWORD dwFlags);
  static void ParseInstanceName(_In_ LPCTSTR szInstanceString, _Inout_ CPDHString& sInstanceName, _Inout_ CPDHString& sParentName, _Out_ DWORD& Index);
  _NODISCARD static CPDHStringArray EnumLogSetNames(_In_ LPCTSTR pszDataSource);
  static void FormatFromRawValue(_In_ DWORD dwCounterType, _In_ DWORD dwFormat, _In_ LONGLONG* pTimeBase, _In_  PPDH_RAW_COUNTER pRawValue1, _In_ PPDH_RAW_COUNTER pRawValue2, _Out_ PPDH_FMT_COUNTERVALUE pFmtValue);
  static void GetDataSourceTimeRange(_In_opt_ LPCTSTR szDataSource, _Out_ LPDWORD pdwNumEntries, _Out_writes_bytes_(*pdwBufferSize) PPDH_TIME_INFO pInfo, _Inout_ LPDWORD pdwBufferSize);
  static void GetDataSourceTimeRangeH(_Inout_opt_ PDH_HLOG hDataSource, _Out_ LPDWORD pdwNumEntries, _Out_writes_bytes_(*pdwBufferSize) PPDH_TIME_INFO pInfo, _Inout_ LPDWORD pdwBufferSize);
  _NODISCARD static DWORD LookupPerfIndexByName(_In_opt_ LPCTSTR szMachineName, _In_ LPCTSTR szName);
  _NODISCARD static CPDHString LookupPerfNameByIndex(_In_opt_ LPCTSTR szMachineName, _In_ DWORD dwIndex);
  static void SetDefaultRealTimeDataSource(_In_ DWORD dwDataSourceId);
  static void ThrowPdhException(_In_ PDH_STATUS Status);
  _NODISCARD static CPDHStringArray ParseMultiSZString(_In_ _NullNull_terminated_ LPCTSTR pszValue);
};

#endif //#ifndef __CPDH_H__
